import os
import subprocess
from datetime import datetime, time
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from aind_data_schema.core.session import Session as AdsSession
from aind_watchdog_service.models.manifest_config import BucketType, ManifestConfig, ModalityConfigs
from aind_watchdog_service.models.watch_config import WatchConfig
from requests.exceptions import HTTPError

from clabe.data_mapper.aind_data_schema import AindDataSchemaSessionDataMapper
from clabe.data_transfer.aind_watchdog import (
    WatchdogDataTransferService,
    WatchdogSettings,
)


@pytest.fixture
def source():
    return Path("source_path")


@pytest.fixture
def aind_data_mapper():
    mapper = MagicMock(spec=AindDataSchemaSessionDataMapper)
    mapper.is_mapped.return_value = True
    mapper.mapped = MagicMock(spec=AdsSession)
    mapper.mapped.experimenter_full_name = ["John Doe"]
    mapper.mapped.subject_id = "12345"
    mapper.mapped.session_start_time = datetime(2023, 1, 1, 10, 0, 0)
    mapper.mapped.data_streams = [MagicMock()]
    mapper.mapped.data_streams[0].stream_modalities = [MagicMock(abbreviation="behavior")]
    return mapper


@pytest.fixture
def settings():
    return WatchdogSettings(
        destination=Path("destination_path"),
        schedule_time=time(hour=20),
        project_name="test_project",
        platform="behavior",
        capsule_id="capsule_id",
        script={"script_key": ["script_value"]},
        s3_bucket=BucketType.PRIVATE,
        mount="mount_path",
        force_cloud_sync=True,
        transfer_endpoint="http://aind-data-transfer-service/api/v1/submit_jobs",
    )


@pytest.fixture
def service(mock_ui_helper, source, settings):
    os.environ["WATCHDOG_EXE"] = "watchdog.exe"
    os.environ["WATCHDOG_CONFIG"] = "watchdog_config.yml"

    service = WatchdogDataTransferService(
        source,
        settings=settings,
        validate=False,
        ui_helper=mock_ui_helper,
    )

    service._manifest_config = ManifestConfig(
        name="test_manifest",
        modalities={"behavior": ["path/to/behavior"], "behavior-videos": ["path/to/behavior-videos"]},
        subject_id=1,
        acquisition_datetime=datetime(2023, 1, 1, 0, 0, 0),
        schemas=["path/to/schema"],
        destination="path/to/destination",
        mount="mount_path",
        processor_full_name="processor_name",
        project_name="test_project",
        schedule_time=settings.schedule_time,
        platform="behavior",
        capsule_id="capsule_id",
        s3_bucket=BucketType.PRIVATE,
        script={"script_key": ["script_value"]},
        force_cloud_sync=True,
        transfer_endpoint="http://aind-data-transfer-service/api/v1/submit_jobs",
    )

    service._watch_config = WatchConfig(
        flag_dir="flag_dir",
        manifest_complete="manifest_complete",
    )

    yield service

    # Cleanup
    if "WATCHDOG_EXE" in os.environ:
        del os.environ["WATCHDOG_EXE"]
    if "WATCHDOG_CONFIG" in os.environ:
        del os.environ["WATCHDOG_CONFIG"]


class TestWatchdogDataTransferService:
    def test_initialization(self, service: WatchdogDataTransferService, settings):
        assert service._settings.destination == settings.destination
        assert service._settings.project_name == settings.project_name
        assert service._settings.schedule_time == settings.schedule_time
        assert service._settings.platform == settings.platform
        assert service._settings.capsule_id == settings.capsule_id
        assert service._settings.script == settings.script
        assert service._settings.s3_bucket == settings.s3_bucket
        assert service._settings.mount == settings.mount
        assert service._settings.force_cloud_sync == settings.force_cloud_sync
        assert service._settings.transfer_endpoint == settings.transfer_endpoint
        assert service.executable_path == Path("watchdog.exe")
        assert service.config_path == Path("watchdog_config.yml")

    def test_aind_session_data_mapper_get(self, service: WatchdogDataTransferService, aind_data_mapper):
        service.with_aind_session_data_mapper(aind_data_mapper)
        assert service.aind_session_data_mapper == aind_data_mapper

    def test_aind_session_data_mapper_get_not_set(self, service: WatchdogDataTransferService):
        service._aind_session_data_mapper = None
        with pytest.raises(ValueError):
            _ = service.aind_session_data_mapper

    def test_with_aind_session_data_mapper(self, service: WatchdogDataTransferService, aind_data_mapper):
        returned_service = service.with_aind_session_data_mapper(aind_data_mapper)
        assert service._aind_session_data_mapper == aind_data_mapper
        assert returned_service == service

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.force_restart", return_value=None)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_service_not_running_restart_success(
        self,
        mock_dump_manifest_config,
        mock_force_restart,
        mock_is_running,
        service: WatchdogDataTransferService,
        aind_data_mapper,
    ):
        mock_is_running.side_effect = [False, True]  # First call returns False, second returns True
        service.with_aind_session_data_mapper(aind_data_mapper)
        service.transfer()
        mock_force_restart.assert_called_once_with(kill_if_running=False)
        mock_dump_manifest_config.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.force_restart",
        side_effect=subprocess.CalledProcessError(1, "cmd"),
    )
    def test_transfer_service_not_running_restart_fail(
        self, mock_force_restart, mock_is_running, service: WatchdogDataTransferService, aind_data_mapper
    ):
        service.with_aind_session_data_mapper(aind_data_mapper)
        with pytest.raises(RuntimeError):
            service.transfer()
        mock_force_restart.assert_called_once_with(kill_if_running=False)

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_data_mapper_not_mapped(
        self, mock_dump_manifest_config, mock_is_running, service: WatchdogDataTransferService, aind_data_mapper
    ):
        aind_data_mapper.is_mapped.return_value = False
        service.with_aind_session_data_mapper(aind_data_mapper)
        with pytest.raises(ValueError):
            service.transfer()
        mock_dump_manifest_config.assert_not_called()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_watch_config_none(
        self, mock_dump_manifest_config, mock_is_running, service: WatchdogDataTransferService, aind_data_mapper
    ):
        service._watch_config = None
        service.with_aind_session_data_mapper(aind_data_mapper)
        with pytest.raises(ValueError):
            service.transfer()
        mock_dump_manifest_config.assert_not_called()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_success(
        self, mock_dump_manifest_config, mock_is_running, service: WatchdogDataTransferService, aind_data_mapper
    ):
        service.with_aind_session_data_mapper(aind_data_mapper)
        service.transfer()
        mock_dump_manifest_config.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=False)
    def test_validate_executable_not_found(self, mock_exists, service: WatchdogDataTransferService):
        with pytest.raises(FileNotFoundError):
            service.validate()

    @patch("clabe.data_transfer.aind_watchdog.Path.exists")
    def test_validate_config_not_found_no_create(self, mock_exists, service: WatchdogDataTransferService):
        mock_exists.side_effect = [True, False]  # executable exists, config does not
        with pytest.raises(FileNotFoundError):
            service.validate(create_config=False)

    @patch("clabe.data_transfer.aind_watchdog.Path.exists")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.create_watch_config")
    def test_validate_config_not_found_create(
        self, mock_create_watch_config, mock_write_yaml, mock_exists, service: WatchdogDataTransferService
    ):
        mock_exists.side_effect = [True, False]  # executable exists, config does not
        mock_create_watch_config.return_value = MagicMock(spec=WatchConfig)
        service.validate(create_config=True)
        mock_create_watch_config.assert_called_once()
        mock_write_yaml.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_service_not_running(
        self, mock_exists, mock_is_running, mock_read_yaml, service: WatchdogDataTransferService
    ):
        assert not service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_invalid_project_name(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        service: WatchdogDataTransferService,
    ):
        assert not service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", side_effect=HTTPError)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_http_error(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        service: WatchdogDataTransferService,
    ):
        with pytest.raises(HTTPError):
            service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_success(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        service: WatchdogDataTransferService,
    ):
        assert service.validate()

    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    @patch("clabe.data_transfer.aind_watchdog.Path.exists")
    def test_create_watch_config_create_dir(self, mock_exists, mock_mkdir):
        mock_exists.side_effect = [False, False]
        watched_dir = Path("test_watched_dir")
        manifest_complete_dir = Path("test_manifest_complete_dir")
        config = WatchdogDataTransferService.create_watch_config(watched_dir, manifest_complete_dir, create_dir=True)
        assert isinstance(config, WatchConfig)
        mock_mkdir.assert_called_with(parents=True, exist_ok=True)
        assert mock_mkdir.call_count == 2

    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=False)
    def test_create_watch_config_no_create_dir(self, mock_exists, mock_mkdir):
        watched_dir = Path("test_watched_dir")
        manifest_complete_dir = Path("test_manifest_complete_dir")
        config = WatchdogDataTransferService.create_watch_config(watched_dir, manifest_complete_dir, create_dir=False)
        assert isinstance(config, WatchConfig)
        mock_mkdir.assert_not_called()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["test_project"],
    )
    def test_is_valid_project_name_valid(self, mock_get_project_names, service: WatchdogDataTransferService):
        assert service.is_valid_project_name()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["other_project"],
    )
    def test_is_valid_project_name_invalid(self, mock_get_project_names, service: WatchdogDataTransferService):
        assert not service.is_valid_project_name()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["other_project"],
    )
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._find_ads_schemas", return_value=[])
    def test_create_manifest_config_from_ads_session_invalid_project_name(
        self, mock_find_ads_schemas, mock_get_project_names, service: WatchdogDataTransferService, aind_data_mapper
    ):
        service._validate_project_name = True
        with pytest.raises(ValueError):
            service.create_manifest_config_from_ads_session(aind_data_mapper.mapped)

    @patch("clabe.data_transfer.aind_watchdog.aind_watchdog_service.models.make_standard_transfer_args")
    def test_add_transfer_service_args_no_jobs(
        self, mock_make_standard_transfer_args, service: WatchdogDataTransferService
    ):
        mock_modality_config = MagicMock(spec=ModalityConfigs)
        mock_modality_config.modality = "behavior-videos"
        mock_transfer_service_args = MagicMock()
        mock_transfer_service_args.upload_jobs = [MagicMock()]
        mock_transfer_service_args.upload_jobs[0].modalities = [mock_modality_config]
        mock_make_standard_transfer_args.return_value = mock_transfer_service_args
        manifest_config = service.add_transfer_service_args(service._manifest_config, jobs=None)
        assert isinstance(manifest_config, ManifestConfig)
        mock_make_standard_transfer_args.assert_called_once()

    def test_add_transfer_service_args_with_callable_jobs(self, service: WatchdogDataTransferService, source):
        def modality_configs_factory(watchdog_service: WatchdogDataTransferService):
            return ModalityConfigs(
                modality="behavior-videos",
                source=Path("gets/replaced/by/service").as_posix(),
                compress_raw_data=True,
                job_settings={"key": "value"},
            )

        assert service._manifest_config is not None

        manifest_config = service.add_transfer_service_args(service._manifest_config, jobs=[modality_configs_factory])

        assert isinstance(manifest_config, ManifestConfig)
        assert manifest_config.transfer_service_args is not None
        assert (
            len(manifest_config.transfer_service_args.upload_jobs[0].modalities) == 2
        )  # The one we add + the default one added by watchdog
        assert (
            manifest_config.transfer_service_args.upload_jobs[0].modalities[-1].modality.abbreviation
            == "behavior-videos"
        )
        assert manifest_config.transfer_service_args.upload_jobs[0].modalities[-1].job_settings == {"key": "value"}

    def test_add_transfer_service_args_with_instance_jobs(self, service: WatchdogDataTransferService):
        modality_configs = ModalityConfigs(
            modality="behavior-videos",
            source=Path("gets/replaced/by/service").as_posix(),
            job_settings={"key": "value"},
        )

        assert service._manifest_config is not None

        manifest_config = service.add_transfer_service_args(service._manifest_config, jobs=[modality_configs])

        assert isinstance(manifest_config, ManifestConfig)
        assert manifest_config.transfer_service_args is not None
        assert (
            len(manifest_config.transfer_service_args.upload_jobs[0].modalities) == 2
        )  # The one we add + the default one added by watchdog
        assert (
            manifest_config.transfer_service_args.upload_jobs[0].modalities[-1].modality.abbreviation
            == "behavior-videos"
        )
        assert manifest_config.transfer_service_args.upload_jobs[0].modalities[-1].job_settings == {"key": "value"}

    @patch("clabe.data_transfer.aind_watchdog.aind_watchdog_service.models.make_standard_transfer_args")
    def test_add_transfer_service_args_with_submit_job_request_kwargs(
        self, mock_make_standard_transfer_args, service: WatchdogDataTransferService
    ):
        mock_transfer_service_args = MagicMock()
        mock_transfer_service_args.model_copy.return_value = mock_transfer_service_args
        mock_make_standard_transfer_args.return_value = mock_transfer_service_args

        submit_kwargs = {"some_key": "some_value"}
        manifest_config = service.add_transfer_service_args(
            service._manifest_config, submit_job_request_kwargs=submit_kwargs
        )

        mock_make_standard_transfer_args.assert_called_once()
        mock_transfer_service_args.model_copy.assert_called_once_with(update=submit_kwargs)
        assert isinstance(manifest_config, ManifestConfig)

    def test_add_transfer_service_args_fail_on_duplicate_modality(self, service: WatchdogDataTransferService, source):
        def modality_configs_factory(watchdog_service: WatchdogDataTransferService):
            return ModalityConfigs(
                modality="behavior-videos",
                source=(Path(watchdog_service._source) / "behavior-videos").as_posix(),
                compress_raw_data=True,
                job_settings={"key": "value"},
            )

        modality_configs = ModalityConfigs(
            modality="behavior-videos",
            source=(Path(service._source) / "behavior-videos").as_posix(),
            job_settings={"key": "value"},
        )

        with pytest.raises(ValueError):
            _ = service.add_transfer_service_args(
                service._manifest_config, jobs=[modality_configs_factory, modality_configs]
            )

    @patch("clabe.data_transfer.aind_watchdog.Path.exists")
    def test_find_ads_schemas_with_existing_schemas(self, mock_exists):
        mock_exists.side_effect = [True] + [
            False
        ] * 50  # First call False, We will prob not have more then 50 modalities...
        source_path = Path("mock_source")
        result = WatchdogDataTransferService._find_ads_schemas(source_path)
        assert len(result) == 1

    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=False)
    def test_find_ads_schemas_no_existing_schemas(self, mock_exists):
        source_path = Path("mock_source")
        result = WatchdogDataTransferService._find_ads_schemas(source_path)
        assert result == []

    @patch("clabe.data_transfer.aind_watchdog.subprocess.check_output")
    def test_is_running(self, mock_check_output, service: WatchdogDataTransferService):
        mock_check_output.return_value = "Image Name                     PID Session Name        Session#    Mem Usage\n========================= ======== ================ =========== =============\nwatchdog.exe                1234 Console                    1    10,000 K\n"
        assert service.is_running()

    @patch("clabe.data_transfer.aind_watchdog.subprocess.check_output")
    def test_is_not_running(self, mock_check_output, service: WatchdogDataTransferService):
        mock_check_output.return_value = "INFO: No tasks are running which match the specified criteria."
        assert not service.is_running()

    @patch("clabe.data_transfer.aind_watchdog.subprocess.run")
    @patch("clabe.data_transfer.aind_watchdog.subprocess.Popen")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running")
    def test_force_restart_kill_if_running(
        self, mock_is_running, mock_popen, mock_run, service: WatchdogDataTransferService
    ):
        mock_is_running.side_effect = [True, False]  # First call returns True, second returns False
        service.force_restart(kill_if_running=True)
        mock_run.assert_called_once_with(
            ["taskkill", "/IM", service.executable_path.name, "/F"], shell=True, check=True
        )
        mock_popen.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.subprocess.run")
    @patch("clabe.data_transfer.aind_watchdog.subprocess.Popen")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    def test_force_restart_no_kill(self, mock_is_running, mock_popen, mock_run, service: WatchdogDataTransferService):
        service.force_restart(kill_if_running=False)
        mock_run.assert_not_called()
        mock_popen.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    def test_dump_manifest_config_custom_path(self, mock_mkdir, mock_write_yaml, service: WatchdogDataTransferService):
        custom_path = Path("custom_path/manifest_test_manifest.yaml")
        result = service.dump_manifest_config(path=custom_path)
        assert result.resolve() == custom_path.resolve()
        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        mock_write_yaml.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    def test_dump_manifest_config_default_path(self, mock_mkdir, mock_write_yaml, service: WatchdogDataTransferService):
        service._watch_config = WatchConfig(flag_dir="flag_dir", manifest_complete="manifest_complete")
        service._manifest_config.name = "test_manifest"
        result = service.dump_manifest_config()
        expected_path = Path("flag_dir/manifest_test_manifest.yaml").resolve()
        assert result.resolve() == expected_path
        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        mock_write_yaml.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    def test_dump_manifest_config_prefix_logic(self, mock_mkdir, mock_write_yaml, service: WatchdogDataTransferService):
        service._watch_config = WatchConfig(flag_dir="flag_dir", manifest_complete="manifest_complete")
        service._manifest_config.name = "test_manifest"
        custom_path = Path("custom_path/my_manifest.yaml")
        result = service.dump_manifest_config(path=custom_path)
        expected_path = Path("custom_path/manifest_my_manifest.yaml").resolve()
        assert result.resolve() == expected_path
        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        mock_write_yaml.assert_called_once()

    def test_dump_manifest_config_no_manifest_config(self, service: WatchdogDataTransferService):
        service._manifest_config = None
        with pytest.raises(ValueError):
            service.dump_manifest_config()

    def test_dump_manifest_config_no_watch_config(self, service: WatchdogDataTransferService):
        service._watch_config = None
        with pytest.raises(ValueError):
            service.dump_manifest_config()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.transfer")
    @patch("clabe.data_transfer.aind_watchdog.BaseLauncher")
    def test_build_runner_callable_aind_session_data_mapper(
        self, MockBaseLauncher, mock_transfer, service: WatchdogDataTransferService, settings
    ):
        mock_launcher = MockBaseLauncher()
        mock_launcher.get_session.return_value = MagicMock(subject="test_subject", session_name="test_session")
        mock_launcher.session_directory = Path("launcher_session_dir")

        def mock_aind_mapper_factory():
            mapper = MagicMock(spec=AindDataSchemaSessionDataMapper)
            mapper.is_mapped.return_value = True
            return mapper

        runner = WatchdogDataTransferService.build_runner(settings, mock_aind_mapper_factory)
        service = runner(mock_launcher)

        assert isinstance(service, WatchdogDataTransferService)
        mock_transfer.assert_called_once()
        assert service._settings.destination == (Path("destination_path") / "test_subject")
        assert service._source == Path("launcher_session_dir")
        assert service._session_name == "test_session"

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.transfer")
    @patch("clabe.data_transfer.aind_watchdog.BaseLauncher")
    def test_build_runner_instance_aind_session_data_mapper(
        self, MockBaseLauncher, mock_transfer, service: WatchdogDataTransferService, settings
    ):
        mock_launcher = MockBaseLauncher()
        mock_launcher.get_session.return_value = MagicMock(subject="test_subject", session_name="test_session")
        mock_launcher.session_directory = Path("launcher_session_dir")

        mock_aind_mapper_instance = MagicMock(spec=AindDataSchemaSessionDataMapper)
        mock_aind_mapper_instance.is_mapped.return_value = True

        runner = WatchdogDataTransferService.build_runner(settings, mock_aind_mapper_instance)
        service = runner(mock_launcher)

        assert isinstance(service, WatchdogDataTransferService)
        mock_transfer.assert_called_once()
        assert service._settings.destination == (Path("destination_path") / "test_subject")
        assert service._source == Path("launcher_session_dir")
        assert service._session_name == "test_session"

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.transfer")
    @patch("clabe.data_transfer.aind_watchdog.BaseLauncher")
    def test_build_runner_data_mapper_not_mapped(self, MockBaseLauncher, mock_transfer, settings):
        mock_launcher = MockBaseLauncher()
        mock_launcher.get_session.return_value = MagicMock(subject="test_subject", session_name="test_session")
        mock_launcher.session_directory = Path("launcher_session_dir")

        mock_aind_mapper_instance = MagicMock(spec=AindDataSchemaSessionDataMapper)
        mock_aind_mapper_instance.is_mapped.return_value = False

        runner = WatchdogDataTransferService.build_runner(settings, mock_aind_mapper_instance)
        with pytest.raises(ValueError):
            runner(mock_launcher)
        mock_transfer.assert_not_called()



import glob
import logging
import os
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Self

from aind_behavior_services.utils import run_bonsai_process
from typing_extensions import override

from ..ui import DefaultUIHelper, UiHelper
from ._base import App

logger = logging.getLogger(__name__)

VISUALIZERS_DIR = "VisualizerLayouts"

if TYPE_CHECKING:
    from ..launcher import BaseLauncher
else:
    BaseLauncher = Any


class BonsaiApp(App):
    """
    A class to manage the execution of Bonsai workflows.

    Attributes:
        executable (os.PathLike): Path to the Bonsai executable.
        workflow (os.PathLike): Path to the Bonsai workflow file.
        is_editor_mode (bool): Whether to run Bonsai in editor mode.
        is_start_flag (bool): Whether to use the start flag when running Bonsai.
        layout (Optional[os.PathLike | str]): Path to the visualizer layout file.
        layout_directory (Optional[os.PathLike]): Directory containing visualizer layouts.
        additional_properties (Optional[Dict[str, str]]): Additional properties to pass to Bonsai.
        cwd (Optional[os.PathLike]): Working directory for the Bonsai process.
        timeout (Optional[float]): Timeout for the Bonsai process.
        print_cmd (bool): Whether to print the command being executed.
        ui_helper (UiHelper): Helper for user interface interactions.
        _result (Optional[subprocess.CompletedProcess]): Result of the Bonsai process execution.
    """

    def __init__(
        self,
        /,
        workflow: os.PathLike,
        executable: os.PathLike = Path("./bonsai/bonsai.exe"),
        is_editor_mode: bool = True,
        is_start_flag: bool = True,
        layout: Optional[os.PathLike] = None,
        layout_dir: Optional[os.PathLike] = None,
        additional_properties: Optional[Dict[str, str]] = None,
        cwd: Optional[os.PathLike] = None,
        timeout: Optional[float] = None,
        print_cmd: bool = False,
        ui_helper: Optional[UiHelper] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the BonsaiApp instance.

        Args:
            workflow (os.PathLike): Path to the Bonsai workflow file.
            executable (os.PathLike): Path to the Bonsai executable. Defaults to "./bonsai/bonsai.exe".
            is_editor_mode (bool): Whether to run Bonsai in editor mode. Defaults to True.
            is_start_flag (bool): Whether to use the start flag. Defaults to True.
            layout (Optional[os.PathLike]): Path to the visualizer layout file. Defaults to None.
            layout_dir (Optional[os.PathLike]): Directory containing visualizer layouts. Defaults to None.
            additional_properties (Optional[Dict[str, str]]): Additional properties for Bonsai. Defaults to None.
            cwd (Optional[os.PathLike]): Working directory for the Bonsai process. Defaults to None.
            timeout (Optional[float]): Timeout for the Bonsai process. Defaults to None.
            print_cmd (bool): Whether to print the command being executed. Defaults to False.
            ui_helper (Optional[UiHelper]): UI helper instance. Defaults to DefaultUIHelper.
            **kwargs: Additional keyword arguments.

        Example:
            ```python
            # Create and run a Bonsai app
            app = BonsaiApp(workflow="workflow.bonsai")
            app.run()
            # Create with custom settings
            app = BonsaiApp(
                workflow="workflow.bonsai",
                is_editor_mode=False,
            )
            ```
        """
        self.executable = Path(executable).resolve()
        self.workflow = Path(workflow).resolve()
        self.is_editor_mode = is_editor_mode
        self.is_start_flag = is_start_flag if not self.is_editor_mode else True
        self.layout = layout
        self.layout_directory = layout_dir
        self.additional_properties = dict(additional_properties) if additional_properties else {}
        self.cwd = cwd
        self.timeout = timeout
        self.print_cmd = print_cmd
        self._result: Optional[subprocess.CompletedProcess] = None
        self.ui_helper = ui_helper if ui_helper is not None else DefaultUIHelper()

    @property
    def result(self) -> subprocess.CompletedProcess:
        """
        Returns the result of the Bonsai process execution.

        Returns:
            subprocess.CompletedProcess: The result of the Bonsai process.

        Raises:
            RuntimeError: If the app has not been run yet.
        """
        if self._result is None:
            raise RuntimeError("The app has not been run yet.")
        return self._result

    @override
    def add_app_settings(self, **kwargs):
        """
        Adds application-specific settings to the additional properties.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            Self: The updated instance of BonsaiApp.
        """

        if self.additional_properties is not None:
            self.additional_properties.update(**kwargs)
        else:
            self.additional_properties = kwargs
        return self

    def validate(self, *args, **kwargs) -> bool:
        """
        Validates the existence of required files and directories.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            bool: True if validation is successful.

        Raises:
            FileNotFoundError: If any required file or directory is missing.
        """
        if not Path(self.executable).exists():
            raise FileNotFoundError(f"Executable not found: {self.executable}")
        if not Path(self.workflow).exists():
            raise FileNotFoundError(f"Workflow file not found: {self.workflow}")
        if self.layout and not Path(self.layout).exists():
            raise FileNotFoundError(f"Layout file not found: {self.layout}")
        if self.layout_directory and not Path(self.layout_directory).exists():
            raise FileNotFoundError(f"Layout directory not found: {self.layout_directory}")
        return True

    @override
    def run(self) -> subprocess.CompletedProcess:
        """
        Runs the Bonsai process.

        Returns:
            subprocess.CompletedProcess: The result of the Bonsai process execution.

        Raises:
            FileNotFoundError: If validation fails.
        """
        self.validate()
        self.prompt_input()

        if self.is_editor_mode:
            logger.warning("Bonsai is running in editor mode. Cannot assert successful completion.")
        logger.info("Bonsai process running...")
        proc = run_bonsai_process(
            workflow_file=self.workflow,
            bonsai_exe=self.executable,
            is_editor_mode=self.is_editor_mode,
            is_start_flag=self.is_start_flag,
            layout=self.layout,
            additional_properties=self.additional_properties,
            cwd=self.cwd,
            timeout=self.timeout,
            print_cmd=self.print_cmd,
        )
        self._result = proc
        logger.info("Bonsai process completed.")
        return proc

    @override
    def output_from_result(self, allow_stderr: Optional[bool]) -> Self:
        """
        Processes the output from the Bonsai process result.

        Args:
            allow_stderr (Optional[bool]): Whether to allow stderr output.

        Returns:
            Self: The updated instance of BonsaiApp.

        Raises:
            subprocess.CalledProcessError: If the process exits with an error.
        """
        proc = self.result
        try:
            proc.check_returncode()
        except subprocess.CalledProcessError as e:
            self._log_process_std_output("Bonsai", proc)
            raise e
        else:
            logger.info("Result from bonsai process is valid.")
            self._log_process_std_output("Bonsai", proc)

            if len(proc.stdout) > 0:
                logger.error("Bonsai process finished with errors.")
                if allow_stderr is None:
                    allow_stderr = self.ui_helper.prompt_yes_no_question("Would you like to see the error message?")
                if allow_stderr is False:
                    raise subprocess.CalledProcessError(1, proc.args)
        return self

    def prompt_visualizer_layout_input(
        self,
        directory: Optional[os.PathLike] = None,
    ) -> Optional[str | os.PathLike]:
        """
        Prompts the user to select a visualizer layout.

        Args:
            directory (Optional[os.PathLike]): Directory containing visualizer layouts.

        Returns:
            Optional[str | os.PathLike]: The selected layout file path.
        """
        if directory is None:
            directory = self.layout_directory
        else:
            directory = Path(os.path.join(directory, VISUALIZERS_DIR, os.environ["COMPUTERNAME"]))

        layout_schemas_path = directory if directory is not None else self.layout_directory
        available_layouts = glob.glob(os.path.join(str(layout_schemas_path), "*.bonsai.layout"))
        picked: Optional[str | os.PathLike] = None
        has_pick = False
        while has_pick is False:
            try:
                picked = self.ui_helper.prompt_pick_from_list(
                    value=available_layouts, prompt="Pick a visualizer layout:"
                )
                picked = picked if picked else ""
                has_pick = True
            except ValueError as e:
                logger.info("Invalid choice. Try again. %s", e)
        self.layout = Path(picked) if picked else None
        return self.layout

    def _log_process_std_output(self, process_name: str, proc: subprocess.CompletedProcess) -> None:
        """
        Logs the standard output and error of a process.

        Args:
            process_name (str): Name of the process.
            proc (subprocess.CompletedProcess): The process result.
        """
        if len(proc.stdout) > 0:
            logger.info("%s full stdout dump: \n%s", process_name, proc.stdout)
        if len(proc.stderr) > 0:
            logger.error("%s full stderr dump: \n%s", process_name, proc.stderr)

    def prompt_input(self, *args, **kwargs):
        """
        Prompts the user for input if required.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Self: The updated instance of BonsaiApp.
        """
        layout_dir = kwargs.pop("layout_directory", None)
        if self.layout is None:
            self.layout = self.prompt_visualizer_layout_input(layout_dir if layout_dir else self.layout_directory)
        return self


class AindBehaviorServicesBonsaiApp(BonsaiApp):
    """
    Specialized Bonsai application for AIND behavior services integration.

    This class extends the base BonsaiApp to provide specific functionality for
    AIND behavior experiments, including automatic configuration of task logic,
    session, and rig paths for the Bonsai workflow.

    Example:
        ```python
        # Create an AIND behavior services Bonsai app
        app = AindBehaviorServicesBonsaiApp(workflow="behavior_workflow.bonsai")
        app.add_app_settings(launcher=my_launcher)
        app.run()
        ```
    """

    def add_app_settings(self, *, launcher: Optional[BaseLauncher] = None, **kwargs) -> Self:
        """
        Adds AIND behavior-specific application settings to the Bonsai workflow.

        This method automatically configures the TaskLogicPath, SessionPath, and RigPath
        properties for the Bonsai workflow based on the launcher's schema models.

        Args:
            launcher: The behavior launcher instance containing schema models
            **kwargs: Additional keyword arguments

        Returns:
            Self: The updated instance

        Raises:
            ValueError: If the required launcher argument is not provided

        Example:
            ```python
            # Add AIND behavior settings
            app.add_app_settings(launcher=my_launcher)
            ```
        """

        if launcher is None:
            raise ValueError("Missing required argument 'launcher'.")

        settings = {
            "TaskLogicPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_task_logic(strict=True), directory=launcher.temp_dir)
            ),
            "SessionPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_session(strict=True), directory=launcher.temp_dir)
            ),
            "RigPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_rig(strict=True), directory=launcher.temp_dir)
            ),
        }
        return super().add_app_settings(**settings)
